/*
 * Copyright (c) 2011 Michael Pippig
 *
 * This file is part of PNFFT.
 *
 * PNFFT is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * PNFFT is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with PNFFT.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include <complex.h>
#include "pnfft.h"
#include "ipnfft.h"
#include "matrix_D.h"

/* wrappers for pfft init and cleanup */
void PNX(init) (void){
  PX(init)();
}


void PNX(cleanup) (void){
  PX(cleanup)();
}


void PNX(local_size_3d)(
    const INT *N, MPI_Comm comm_cart,
    INT *local_N, INT *local_N_start,
    R *lower_border, R *upper_border
    )
{
  const int d=3;

  PNX(local_size_adv)(
      d, N, comm_cart,
      local_N, local_N_start, lower_border, upper_border);
}


void PNX(init_3d)(
    PNX(plan) *ths, const INT *N,
    INT local_M,
    MPI_Comm comm_cart
    )
{
  const int d=3;
  unsigned nfft_flags, fftw_flags;

  nfft_flags = PNFFT_MALLOC_X | PNFFT_MALLOC_F_HAT | PNFFT_MALLOC_F | PNFFT_FFT_INIT;
  fftw_flags = FFTW_MEASURE;

  PNX(init_adv)(
      ths, d, N, local_M,
      nfft_flags, fftw_flags, comm_cart);
}


void PNX(init_nodes)(
    PNX(plan) *ths, INT local_M,
    unsigned nfft_flags, unsigned nfft_finalize_flags
    )
{
  /* free mem and adjust nfft_flags, compute_flags */
  PNX(free_x)(ths, nfft_finalize_flags);
  PNX(free_f)(ths, nfft_finalize_flags);
  PNX(free_grad_f)(ths, nfft_finalize_flags);

  /* allocate mem and adjust nfft_flags, compute_flags */
  ths->local_M = local_M;
  PNX(malloc_x)(ths, nfft_flags);
  PNX(malloc_f)(ths, nfft_flags);
  PNX(malloc_grad_f)(ths, nfft_flags);
}


/* parallel 3dNFFT with Gaussian window function */
void PNX(trafo)(
    PNX(plan) *ths
    )
{
  if(ths==NULL){
    PX(fprintf)(MPI_COMM_WORLD, stderr, "!!! Error: Can not execute PNFFT Plan == NULL !!!\n");
    return;
  }

  ths->timer_trafo.whole -= MPI_Wtime();

  /* multiplication with matrix D */
  ths->timer_trafo.matrix_D -= MPI_Wtime();
  PNX(trafo_D)(ths);
  ths->timer_trafo.matrix_D += MPI_Wtime();

  /* multiplication with matrix F */
  ths->timer_trafo.matrix_F -= MPI_Wtime();
  PNX(trafo_F)(ths);
  ths->timer_trafo.matrix_F += MPI_Wtime();

  /* multiplication with matrix B */
  ths->timer_trafo.matrix_B -= MPI_Wtime();
  PNX(trafo_B)(ths);
  ths->timer_trafo.matrix_B += MPI_Wtime();
 
  ths->timer_trafo.iter++;
  ths->timer_trafo.whole += MPI_Wtime();
}


void PNX(adj)(
    PNX(plan) *ths
    )
{
  if(ths==NULL){
    PX(fprintf)(MPI_COMM_WORLD, stderr, "!!! Error: Can not execute PNFFT Plan == NULL !!!\n");
    return;
  }

  ths->timer_adj.whole -= MPI_Wtime();

  /* multiplication with matrix B^T */
  ths->timer_adj.matrix_B -= MPI_Wtime();
  PNX(adjoint_B)(ths);
  ths->timer_adj.matrix_B += MPI_Wtime();

  /* multiplication with matrix F^H */
  ths->timer_adj.matrix_F -= MPI_Wtime();
  PNX(adjoint_F)(ths);
  ths->timer_adj.matrix_F += MPI_Wtime();

  /* multiplication with matrix D */
  ths->timer_adj.matrix_D -= MPI_Wtime();
  PNX(adjoint_D)(ths);
  ths->timer_adj.matrix_D += MPI_Wtime();

  ths->timer_adj.iter++;
  ths->timer_adj.whole += MPI_Wtime();
}


void PNX(finalize)(
    PNX(plan) *ths, unsigned nfft_finalize_flags
    )
{
  if((nfft_finalize_flags & PNFFT_FREE_F_HAT) && (ths->f_hat != NULL))
    PNX(free)(ths->f_hat);
  if((nfft_finalize_flags & PNFFT_FREE_GRAD_F) && (ths->grad_f != NULL))
    PNX(free)(ths->grad_f);
  if((nfft_finalize_flags & PNFFT_FREE_F) && (ths->f != NULL))
    PNX(free)(ths->f);
  if((nfft_finalize_flags & PNFFT_FREE_X) && (ths->x != NULL))
    PNX(free)(ths->x);

  if(ths->nfft_flags & PNFFT_FFT_INIT)
  {
    PX(destroy_plan)(ths->pfft_forw);
    PX(destroy_plan)(ths->pfft_back);
    PX(destroy_gcplan)(ths->gcplan);

    PNX(free)(ths->g1);
    if(ths->nfft_flags & PNFFT_FFT_OUT_OF_PLACE)
      PNX(free)(ths->g2);
  }

  PNX(free)(ths->x_max);
  PNX(free)(ths->sigma);
  PNX(free)(ths->n);
  PNX(free)(ths->no);
  PNX(free)(ths->N);

  MPI_Comm_free(&(ths->comm_cart));

  /* finalize window specific parameters */
  PNX(free)(ths->b);
  if(ths->exp_const != NULL)
    PNX(free)(ths->exp_const);
}


void PNX(init_f_hat_3d)(
    const INT *N, const INT *local_N, const INT *local_N_start,
    C *data
    )
{
  PX(init_input_c2c_3d)(N, local_N, local_N_start,
      data);
}


void PNX(init_x_3d)(
    const R *lo, const R *up, INT loc_M,
    R *x
    )
{
  R x_max[3] = {0.5,0.5,0.5};

  PNX(init_x_3d_adv)(lo, up, x_max, loc_M,
      x);
}


static void print_complex_vector(
    R *data, INT N
    )
{
  for(INT l=0; l<N; l++){
    if(l%4 == 0)
      printf("\n%4td.", l/4);
    printf(" %.2e+%.2ei,", data[2*l], data[2*l+1]);
  }
  printf("\n");
}


void PNX(vpr_complex)(
    C *data, INT N,
    const char *name, MPI_Comm comm
    )
{
  int size, myrank;

  if(N < 1)
    return;

  MPI_Comm_size(comm, &size);
  MPI_Comm_rank(comm, &myrank);
  
  fflush(stdout);
  MPI_Barrier(comm);
  for(int t=0; t<size; t++){
    if(t==myrank){
      printf("\nRank %d, %s", myrank, name);
      print_complex_vector((R*) data, N);
      fflush(stdout);
    }
    MPI_Barrier(comm);
  }
}


void PNX(vpr_real)(
    R *data, INT N,
    const char *name, MPI_Comm comm
    )
{
  int size, myrank;

  if(N < 1)
    return;
  
  MPI_Comm_size(comm, &size);
  MPI_Comm_rank(comm, &myrank);
  
  fflush(stdout);
  MPI_Barrier(comm);
  for(int t=0; t<size; t++){
    if(t==myrank){
      printf("\nRank %d, %s", myrank, name);
      for(INT l=0; l<N; l++){
        if(l%8 == 0)
          printf("\n%4td.", l/8);
        printf(" %e,", data[l]);
      }
      printf("\n");
      fflush(stdout);
    }
    MPI_Barrier(comm);
  }
}


void PNX(apr_complex_3d)(
     C *data, INT *local_N, INT *local_N_start,
     const char *name, MPI_Comm comm
     )
{
  PX(apr_complex_3d)(
     data, local_N, local_N_start, name, comm);
}
