/*
 * Copyright (c) 2003, 2007-8 Matteo Frigo
 * Copyright (c) 2003, 2007-8 Massachusetts Institute of Technology
 * Copyright (c) 2002, 2009 Jens Keiner, Stefan Kunis, Daniel Potts
 * Copyright (c) 2010, 2011 Michael Pippig
 *
 * This file is part of PNFFT.
 *
 * PNFFT is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * PNFFT is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with PNFFT.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#ifndef PNFFT_H
#define PNFFT_H 1

#include <math.h>
#include <mpi.h>
#include <pfft.h>

#ifdef __cplusplus
  extern "C"
  {
#endif /* __cplusplus */


#define PNFFT_EXTERN extern
#define PNFFT_CONCAT(prefix, name) prefix ## name


  /*
    huge second-order macro that defines prototypes for all API
    functions.  We expand this macro for each supported precision

    PNX: PNFFT name-mangling macro (parallel NFFT)
    PX: PFFT name-mangling macro (parallel FFT)
    X : FFTW name-mangling macro (serial FFT)
    R : real data type
    C : complex data type
  */
#define PNFFT_DEFINE_API(PNX, PX, X, R, C, INT)                                         \
                                                                                        \
  typedef struct{                                                                       \
    int iter;                                                                           \
    double whole;                                                                       \
    double loop_B;                                                                      \
    double sort_nodes;                                                                  \
    double gcells;                                                                      \
    double matrix_B;                                                                    \
    double matrix_F;                                                                    \
    double matrix_D;                                                                    \
  } PNX(timer);                                                                         \
                                                                                        \
  typedef struct{                                                                       \
    INT N_total;                /**< Total number of Fourier coefficients            */ \
    C *f_hat;                   /**< Vector of Fourier coefficients                  */ \
    C *f;                       /**< Vector of samples                               */ \
    C *grad_f;                  /**< Vector of gradients                             */ \
    R *x;                       /**< Nodes in time/spatial domain                    */ \
                                                                                        \
    int d;                      /**< Dimension, rank                                 */ \
    INT *N;                     /**< Multi bandwidth                                 */ \
    R *sigma;                   /**< Oversampling-factor                             */ \
    INT *n;                     /**< FFT length, equal to sigma*N                    */ \
    INT *no;                    /**< FFT output length                               */ \
    INT n_total;                /**< Total size of FFTW                              */ \
    int m;                      /**< Cut-off parameter of the window function        */ \
    R *x_max;                   /**< Upper border for nodes in time/spatial domain   */ \
                                                                                        \
    R *b;                       /**< Shape parameter of Gaussian window function     */ \
    R *exp_const;               /**< Precomputed values for Fast Gaussian window     */ \
                                                                                        \
    unsigned nfft_flags;        /**< Flags for precomputation, (de)allocation,          \
                                     and FFTW usage                                  */ \
    unsigned compute_flags;     /**< Flags for choice of NFFT results                */ \
    unsigned fftw_flags;        /**< Flags for the FFTW                              */ \
                                                                                        \
    /* internal*/                                                                       \
    PX(plan)   pfft_forw;       /**< Forward PFFT plan                               */ \
    PX(plan)   pfft_back;       /**< Backward PFFT plan                              */ \
    PX(gcplan) gcplan;          /**< PFFT Ghostcell plan                             */ \
                                                                                        \
    C *g1;                      /**< Input of PFFT                                   */ \
    C *g2;                      /**< Output of PFFT                                  */ \
                                                                                        \
    int cutoff;                 /**< cutoff range                                    */ \
    INT local_M;                /**< Number of local nodes                           */ \
                                                                                        \
    MPI_Comm comm_cart;      /**< Two-dimensional Cartesian communicator          */ \
    int np[3];                  /**< Size of Cartesian communicator                  */ \
                                                                                        \
    PNX(timer) timer_trafo;     /**< Saves time measurements during PNFFT            */ \
    PNX(timer) timer_adj;       /**< Saves time measurements during adjoint PNFFT    */ \
  } PNX(plan);                                                                          \
                                                                                        \
  PNFFT_EXTERN int PNX(create_procmesh_2d)(                                             \
      MPI_Comm comm, int np0, int np1, MPI_Comm *comm_cart_2d);                         \
  PNFFT_EXTERN int PNX(create_procmesh)(                                                \
      int rnk, MPI_Comm comm, const int *np, MPI_Comm *comm_cart);                      \
                                                                                        \
  PNFFT_EXTERN void PNX(local_size_3d)(                                                 \
      const INT *N, MPI_Comm comm_cart,                                                 \
      INT *local_N, INT *local_N_start,                                                 \
      R *lower_border, R *upper_border);                                                \
  PNFFT_EXTERN void PNX(local_size_adv)(                                                \
      int d, const INT *N, MPI_Comm comm_cart,                                          \
      INT *local_N, INT *local_N_start,                                                 \
      R *lower_border, R *upper_border);                                                \
  PNFFT_EXTERN void PNX(local_size_guru)(                                               \
      int d, const INT *N, const INT *n, const R *x_max, int m,                         \
      MPI_Comm comm_cart,                                                               \
      INT *local_N, INT *local_N_start,                                                 \
      R *lower_border, R *upper_border);                                                \
                                                                                        \
  PNFFT_EXTERN void PNX(init_3d)(                                                       \
      PNX(plan) *ths, const INT *N, INT local_M,                                        \
      MPI_Comm comm_cart);                                                              \
  PNFFT_EXTERN void PNX(init_adv)(                                                      \
      PNX(plan) *ths, int d, const INT *N, INT local_M,                                 \
      unsigned nfft_flags, unsigned fftw_flags, MPI_Comm comm_cart);                    \
  PNFFT_EXTERN void PNX(init_guru)(                                                     \
        PNX(plan) *ths, int d,                                                          \
        const INT *N, const INT *n, const R *x_max,                                     \
        INT local_M, int m,                                                             \
        unsigned nfft_flags, unsigned fftw_flags,                                       \
        MPI_Comm comm_cart);                                                            \
                                                                                        \
  PNFFT_EXTERN void PNX(init_nodes)(                                                    \
      PNX(plan) *ths, INT local_M,                                                      \
      unsigned nfft_flags, unsigned nfft_finalize_flags);                               \
                                                                                        \
  PNFFT_EXTERN void PNX(finalize)(                                                      \
        PNX(plan) *ths, unsigned nfft_finalize_flags);                                  \
                                                                                        \
  PNFFT_EXTERN void PNX(trafo)(                                                         \
      PNX(plan) *ths);                                                                  \
  PNFFT_EXTERN void PNX(adj)(                                                           \
      PNX(plan) *ths);                                                                  \
                                                                                        \
  PNFFT_EXTERN void PNX(init)(                                                          \
      void);                                                                            \
  PNFFT_EXTERN void PNX(cleanup)(                                                       \
      void);                                                                            \
                                                                                        \
  PNFFT_EXTERN void *PNX(malloc)(size_t n);					        \
  PNFFT_EXTERN R *PNX(alloc_real)(size_t n);					        \
  PNFFT_EXTERN C *PNX(alloc_complex)(size_t n);				                \
  PNFFT_EXTERN void PNX(free)(void *p);					                \
                                                                                        \
  PNFFT_EXTERN void PNX(init_f_hat_3d)(                                                 \
      const INT *N, const INT *local_N,                                                 \
      const INT *local_N_start,                                                         \
      C *data);                                                                         \
  PNFFT_EXTERN void PNX(init_x_3d)(                                                     \
    const R *lo, const R *up, INT loc_M,                                                \
    R *x);                                                                              \
  PNFFT_EXTERN void PNX(init_x_3d_adv)(                                                 \
    const R *lo, const R *up, const R *x_max, INT loc_M,                                \
    R *x);                                                                              \
                                                                                        \
  PNFFT_EXTERN void PNX(vpr_complex)(                                                   \
    C *data, INT N, const char *name, MPI_Comm comm);                                   \
  PNFFT_EXTERN void PNX(vpr_real)(                                                      \
    R *data, INT N, const char *name, MPI_Comm comm);                                   \
  PNFFT_EXTERN void PNX(apr_complex_3d)(                                                \
    C *data, INT *local_N, INT *local_N_start,                                          \
    const char *name, MPI_Comm comm);                                                   \
                                                                                        \
  PNFFT_EXTERN void PNX(convert_timer2vec)(                                             \
      const PNX(timer) *timer, double *times);                                          \
  PNFFT_EXTERN void PNX(convert_vec2timer)(                                             \
    const double *times, PNX(timer) *timer);                                            \
  PNFFT_EXTERN void PNX(average_timer)(                                                 \
    PNX(timer) *timer);                                                                 \
  PNFFT_EXTERN void PNX(copy_timer)(                                                    \
    const PNX(timer) *orig, PNX(timer) *copy);                                          \
  PNFFT_EXTERN void PNX(reduce_max_timer)(                                              \
    MPI_Comm comm, const PNX(timer) *timer, PNX(timer) *timer_max);                     \
  PNFFT_EXTERN void PNX(add_timers)(                                                    \
    const PNX(timer) *sum1, const PNX(timer) *sum2, PNX(timer) *res);                   \
                                                                                        \
  PNFFT_EXTERN void PNX(get_timer_trafo)(                                               \
    PNX(plan) *ths, PNX(timer) *timer);                                                 \
  PNFFT_EXTERN void PNX(get_timer_adj)(                                                 \
    PNX(plan) *ths, PNX(timer) *timer);                                                 \
  PNFFT_EXTERN void PNX(reset_timer)(                                                   \
      PNX(plan) *ths);                                                                  \
  PNFFT_EXTERN void PNX(print_average_timer)(                                           \
      PNX(plan) *ths, MPI_Comm comm);                                                   \
  PNFFT_EXTERN void PNX(print_average_timer_adv)(                                       \
      PNX(plan) *ths, MPI_Comm comm);                                                   \
  PNFFT_EXTERN void PNX(write_average_timer)(                                           \
      PNX(plan) *ths, const char *name, MPI_Comm comm);                                 \
  PNFFT_EXTERN void PNX(write_average_timer_adv)(                                       \
      PNX(plan) *ths, const char *name, MPI_Comm comm);                                 \
                                                                                        \
  PNFFT_EXTERN void PNX(get_args)(                                                      \
      int argc, char **argv, const char *name,                                          \
      int neededArgs, unsigned type,                                                    \
      void *parameter);
    

typedef pfft_complex pnfft_complex;
typedef pfftf_complex pnfftf_complex;
typedef pfftl_complex pnfftl_complex;

#define PNFFT_MANGLE_DOUBLE(name) PNFFT_CONCAT(pnfft_, name)
#define PNFFT_MANGLE_FLOAT(name) PNFFT_CONCAT(pnfftf_, name)
#define PNFFT_MANGLE_LONG_DOUBLE(name) PNFFT_CONCAT(pnfftl_, name)

PNFFT_DEFINE_API(
    PNFFT_MANGLE_DOUBLE, PFFT_MANGLE_DOUBLE, FFTW_MANGLE_DOUBLE,
    double, pnfft_complex, ptrdiff_t)
PNFFT_DEFINE_API(
    PNFFT_MANGLE_FLOAT, PFFT_MANGLE_FLOAT, FFTW_MANGLE_FLOAT,
    float, pnfftf_complex, ptrdiff_t)
PNFFT_DEFINE_API(
    PNFFT_MANGLE_LONG_DOUBLE, PFFT_MANGLE_LONG_DOUBLE, FFTW_MANGLE_LONG_DOUBLE,
    long double, pnfftl_complex, ptrdiff_t)




#ifndef PNFFT_PI
#define PNFFT_PI           3.14159265358979323846
#endif

#define PNFFT_PRE_PHI_HUT      (1U<< 0)
#define PNFFT_FG_PSI           (1U<< 1)
#define PNFFT_PRE_LIN_PSI      (1U<< 2)
#define PNFFT_PRE_FG_PSI       (1U<< 3)
#define PNFFT_PRE_PSI          (1U<< 4)
#define PNFFT_PRE_FULL_PSI     (1U<< 5)

#define PNFFT_MALLOC_X         (1U<< 6)
#define PNFFT_MALLOC_F_HAT     (1U<< 7)
#define PNFFT_MALLOC_F         (1U<< 8)
#define PNFFT_MALLOC_GRAD_F    (1U<< 9)
#define PNFFT_FFT_OUT_OF_PLACE (1U<< 10)
#define PNFFT_FFT_INIT         (1U<< 11)
#define PNFFT_SORT_NODES       (1U<< 12)

/* default window function is Kaiser-Bessel */
#define PNFFT_WINDOW_KAISER_BESSEL (0U)
#define PNFFT_WINDOW_GAUSSIAN  (1U<< 13)

#define PNFFT_PRE_ONE_PSI (PNFFT_PRE_LIN_PSI| PNFFT_PRE_FG_PSI| PNFFT_PRE_PSI| PNFFT_PRE_FULL_PSI)

#define PNFFT_FREE_X           PNFFT_MALLOC_X
#define PNFFT_FREE_F_HAT       PNFFT_MALLOC_F_HAT
#define PNFFT_FREE_F           PNFFT_MALLOC_F
#define PNFFT_FREE_GRAD_F      PNFFT_MALLOC_GRAD_F

#define PNFFT_COMPUTE_F        PNFFT_MALLOC_F
#define PNFFT_COMPUTE_GRAD_F   PNFFT_MALLOC_GRAD_F

#define PNFFT_INT            PFFT_INT
#define PNFFT_PTRDIFF_T      PFFT_PTRDIFF_T
#define PNFFT_FLOAT          PFFT_FLOAT
#define PNFFT_DOUBLE         PFFT_DOUBLE
#define PNFFT_UNSIGNED       PFFT_UNSIGNED



#ifdef __cplusplus
  }  /* extern "C" */
#endif /* __cplusplus */

#endif /* PNFFT_H */

