/*
 * Copyright (c) 2011 Michael Pippig
 *
 * This file is part of PNFFT.
 *
 * PNFFT is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * PNFFT is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with PNFFT.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include <complex.h>
#include "pnfft.h"
#include "ipnfft.h"
#include "matrix_D.h"
#include "bessel_i0.h"

static void convolution_with_kaiser_bessel_trafo(
    const C *f_hat,
    const INT *n, const INT *no,
    const INT *local_N, const INT *local_N_start,
    const R *b, int m,
    C *g_hat);
static void convolution_with_gaussian_trafo(
    const C *f_hat,
    const INT *n, const INT *no,
    const INT *local_N, const INT *local_N_start,
    const R *b,
    C *g_hat);
static void convolution_with_kaiser_bessel_adjoint(
    const C *g_hat,
    const INT *n, const INT *no,
    const INT *local_N, const INT *local_N_start,
    const R *b, int m,
    C *f_hat);
static void convolution_with_gaussian_adjoint(
    const C *g_hat,
    const INT *n, const INT *no,
    const INT *local_N, const INT *local_N_start,
    const R *b,
    C *f_hat);


void PNX(local_size_D)(
    const INT *N, const INT *n, const INT *no,
    MPI_Comm comm_cart,
    INT *local_N, INT *local_N_start
    )
{
  INT local_no[3], local_no_start[3];

  PNX(local_size_internal)(N, n, no, comm_cart,
      local_N, local_N_start, local_no, local_no_start);
}


void PNX(trafo_D)(
    PNX(plan) *ths
    )
{
  INT local_N[3], local_N_start[3];
  
  PNX(local_size_D)(ths->N, ths->n, ths->no, ths->comm_cart,
      local_N, local_N_start);

  if(ths->nfft_flags & PNFFT_WINDOW_GAUSSIAN)
    convolution_with_gaussian_trafo(
        ths->f_hat, ths->n, ths->no, local_N, local_N_start, ths->b,
        ths->g1);
  else
    convolution_with_kaiser_bessel_trafo(
        ths->f_hat, ths->n, ths->no, local_N, local_N_start, ths->b, ths->m,
        ths->g1);
}


void PNX(adjoint_D)(
    PNX(plan) *ths
    )
{
  INT local_N[3], local_N_start[3];
  
  PNX(local_size_D)(ths->N, ths->n, ths->no, ths->comm_cart,
      local_N, local_N_start);
  
  if(ths->nfft_flags & PNFFT_WINDOW_GAUSSIAN)
    convolution_with_gaussian_adjoint(
        ths->g1, ths->n, ths->no, local_N, local_N_start, ths->b,
        ths->f_hat);
  else
    convolution_with_kaiser_bessel_adjoint(
        ths->g1, ths->n, ths->no, local_N, local_N_start, ths->b, ths->m,
        ths->f_hat);
}


static void convolution_with_kaiser_bessel_trafo(
    const C *f_hat,
    const INT *n, const INT *no,
    const INT *local_N, const INT *local_N_start,
    const R *b, int m,
    C *g_hat
    )
{
  INT k0, k1, k2, k=0;

  /* g_hat is transposed N1 x N2 x N0 */
  for(k1=local_N_start[1]; k1<local_N_start[1] + local_N[1]; k1++)
    for(k2=local_N_start[2]; k2<local_N_start[2] + local_N[2]; k2++)
      for(k0=local_N_start[0]; k0<local_N_start[0] + local_N[0]; k0++, k++)
        g_hat[k] = f_hat[k]
          * MACRO_INV_PHI_HAT_KAISER(k0, n[0], b[0], m)
          * MACRO_INV_PHI_HAT_KAISER(k1, n[1], b[1], m)
          * MACRO_INV_PHI_HAT_KAISER(k2, n[2], b[2], m)
          * cexp(-FFTW_FORWARD * PNFFT_PI * I * ( k0*no[0]/((R)n[0]) + k1*no[1]/((R) n[1]) + k2*no[2]/((R) n[2]) ) );
}

static void convolution_with_gaussian_trafo(
    const C *f_hat,
    const INT *n, const INT *no,
    const INT *local_N, const INT *local_N_start,
    const R *b,
    C *g_hat
    )
{
  INT k0, k1, k2, k=0;

  /* g_hat is transposed N1 x N2 x N0 */
  for(k1=local_N_start[1]; k1<local_N_start[1] + local_N[1]; k1++)
    for(k2=local_N_start[2]; k2<local_N_start[2] + local_N[2]; k2++)
      for(k0=local_N_start[0]; k0<local_N_start[0] + local_N[0]; k0++, k++)
        g_hat[k] = f_hat[k]
          * MACRO_INV_PHI_HAT_GAUSS(k0, n[0], b[0])
          * MACRO_INV_PHI_HAT_GAUSS(k1, n[1], b[1])
          * MACRO_INV_PHI_HAT_GAUSS(k2, n[2], b[2])
          * cexp(-FFTW_FORWARD * PNFFT_PI * I * ( k0*no[0]/((R)n[0]) + k1*no[1]/((R) n[1]) + k2*no[2]/((R) n[2]) ) );
}

static void convolution_with_kaiser_bessel_adjoint(
    const C *g_hat,
    const INT *n, const INT *no,
    const INT *local_N, const INT *local_N_start,
    const R *b, int m,
    C *f_hat
    )
{
  INT k0, k1, k2, k=0;

  /* g_hat is transposed N1 x N2 x N0 */
  for(k1=local_N_start[1]; k1<local_N_start[1] + local_N[1]; k1++)
    for(k2=local_N_start[2]; k2<local_N_start[2] + local_N[2]; k2++)
      for(k0=local_N_start[0]; k0<local_N_start[0] + local_N[0]; k0++, k++)
        f_hat[k] = g_hat[k]
          * MACRO_INV_PHI_HAT_KAISER(k0, n[0], b[0], m)
          * MACRO_INV_PHI_HAT_KAISER(k1, n[1], b[1], m)
          * MACRO_INV_PHI_HAT_KAISER(k2, n[2], b[2], m)
          * cexp(-FFTW_BACKWARD * PNFFT_PI * I * ( k0*no[0]/((R)n[0]) + k1*no[1]/((R) n[1]) + k2*no[2]/((R) n[2]) ) );
}

static void convolution_with_gaussian_adjoint(
    const C *g_hat,
    const INT *n, const INT *no,
    const INT *local_N, const INT *local_N_start,
    const R *b,
    C *f_hat
    )
{
  INT k0, k1, k2, k=0;

  /* g_hat is transposed N1 x N2 x N0 */
  for(k1=local_N_start[1]; k1<local_N_start[1] + local_N[1]; k1++)
    for(k2=local_N_start[2]; k2<local_N_start[2] + local_N[2]; k2++)
      for(k0=local_N_start[0]; k0<local_N_start[0] + local_N[0]; k0++, k++)
        f_hat[k] = g_hat[k]
          * MACRO_INV_PHI_HAT_GAUSS(k0, n[0], b[0])
          * MACRO_INV_PHI_HAT_GAUSS(k1, n[1], b[1])
          * MACRO_INV_PHI_HAT_GAUSS(k2, n[2], b[2])
          * cexp(-FFTW_BACKWARD * PNFFT_PI * I * ( k0*no[0]/((R)n[0]) + k1*no[1]/((R) n[1]) + k2*no[2]/((R) n[2]) ) );
}

