/*
 * Copyright (c) 2011 Michael Pippig
 *
 * This file is part of PNFFT.
 *
 * PNFFT is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * PNFFT is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with PNFFT.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include <complex.h>
#include "pnfft.h"
#include "ipnfft.h"

static void get_size_gcells(
    int m, int cutoff,
    INT *gcells_below, INT *gcells_above);
static void lowest_summation_index(
    const INT *n, int m, const R *x,
    const INT *local_no_start, const INT *gcells_below,
    INT *u_j);
static void local_array_size(
    const INT *local_n, const INT *gcells_below, const INT *gcells_above,
    INT *local_ngc);
static void pre_psi_tensor(
    const INT *n, const R *b, int m, int cutoff,
    const R *x, const R *exp_const, unsigned nfft_flags,
    R *pre_psi);
static void pre_psi_tensor_kaiser_bessel(
    const INT *n, const R *b, int m, int cutoff,
    const R *x,
    R *pre_psi);
static void pre_psi_tensor_gaussian(
    const INT *n, const R *b, int m, int cutoff,
    const R *x, const R *exp_const,
    R *fg_psi);
static void pre_psi_derive_tensor(
    const INT *n, const R *b, int m, int cutoff,
    const R *x, const R *pre_psi, unsigned nfft_flags,
    R *pre_psi_derive);
static void pre_psi_derive_tensor_kaiser_bessel(
    const INT *n, const R *b, int m, int cutoff,
    const R *x, const R *pre_psi,
    R *pre_psi_derive);
static void pre_psi_derive_tensor_gaussian(
    const INT *n, const R *b, int m, int cutoff,
    const R *x, const R *fg_psi,
    R *fg_psi_derive);
static void sort_nodes_for_better_cache_handle(
    const INT *n, int m, INT local_M, const R *x,
    INT *sort);
static void project_node_to_grid(
    const INT *n, int m, const R *x,
    INT *u_j);
static void get_mpi_cart_dims(
    MPI_Comm comm_cart,
    int *dims);
static int compare_INT(
    const void* a, const void* b);

static R kaiser_bessel_1d(
    R x, INT n, R b, int m);
static R kaiser_bessel_derivative_1d(
    R x, INT n, R b, int m, R psi);


/* The local borders of the FFT output imply the box of nodes
 * that can be handled by the current process. If x_max < 0.5,
 * the FFT output size 'no' includes some extra elements at the
 * beginning and the end (the number of extra elements depends on
 * the interpolation order 'm').
 * Therefore, we must explicitly check, if the borders are not smaller
 * then -x_max and not bigger then x_max.
 */
void PNX(node_borders)(
    const INT *n,
    const INT* local_no, const INT *local_no_start,
    const R* x_max,
    R *lo, R *up
    )
{
  for(int t=0; t<3; t++){
    lo[t] = local_no_start[t] / ((R) n[t]);    
    up[t] = (local_no_start[t] + local_no[t]) / ((R) n[t]);

    /* assure, that borders lie in [-x_max,x_max] */
    if(lo[t] < -x_max[t])
      lo[t] = -x_max[t];
    if(lo[t] > x_max[t])
      lo[t] = x_max[t];
    if(up[t] < -x_max[t])
      up[t] = -x_max[t];
    if(up[t] > x_max[t])
      up[t] = x_max[t];
  }
}

void PNX(local_size_B)(
    const INT *N, const INT *n, const INT *no,
    MPI_Comm comm_cart,
    INT *local_no, INT *local_no_start
    )
{
  INT local_N[3], local_N_start[3];

  PNX(local_size_internal)(N, n, no, comm_cart,
      local_N, local_N_start, local_no, local_no_start);
}

void PNX(local_size_internal)(
    const INT *N, const INT *n, const INT *no,
    MPI_Comm comm_cart,
    INT *local_N, INT *local_N_start,
    INT *local_no, INT *local_no_start
    )
{
  INT alloc_local, howmany = 1;

  alloc_local = PX(local_size_many_dft)(3, n, N, no, howmany,
      PFFT_DEFAULT_BLOCKS, PFFT_DEFAULT_BLOCKS, comm_cart, PFFT_TRANSPOSED_IN,
      local_N, local_N_start, local_no, local_no_start);

  for(int t=0; t<3; t++){
    local_N_start[t] -= N[t]/2;
    local_no_start[t] -= no[t]/2;
  }
}



/* N - size of NFFT
 * n - oversampled FFT size
 * no - FFT output size (if nodes are only in a subset the array) */
void PNX(init_internal)(
    PNX(plan) *ths, int d, const INT *N, const INT *n, const INT *no,
    INT local_M, int m,
    unsigned nfft_flags, unsigned fftw_flags,
    MPI_Comm comm_cart
    )
{
  /* force Gaussian window for debugging */
//  nfft_flags |= PNFFT_WINDOW_GAUSSIAN;

  INT howmany = 1;
  INT alloc_local_data_in, alloc_local_gc, alloc_local_f_hat;
  INT gcells_below[3], gcells_above[3];
  INT local_N[3], local_N_start[3], local_no[3], local_no_start[3];
  INT local_ngc[3], local_gc_start[3];

  ths->d = d;
  ths->m= m;
  ths->local_M = local_M;

  ths->N = (INT*) PNX(malloc)(sizeof(INT) * (size_t) d);
  ths->n = (INT*) PNX(malloc)(sizeof(INT) * (size_t) d);
  ths->no= (INT*) PNX(malloc)(sizeof(INT) * (size_t) d);
  for(int t=0; t<d; t++){
    ths->N[t]= N[t];
    ths->n[t]= n[t];
    ths->no[t]= no[t];
  }

  ths->nfft_flags = nfft_flags;
  ths->fftw_flags = fftw_flags;

  MPI_Comm_dup(comm_cart, &(ths->comm_cart));
  get_mpi_cart_dims(comm_cart, ths->np);
  
  ths->cutoff = 2*m+2;
  ths->N_total = ths->n_total = 1;
  for(int t=0; t<d; t++){
    ths->N_total *= N[t];
    ths->n_total *= n[t];
  }
  /* x_max is filled in init_guru */
  ths->x_max = (R*) PNX(malloc)(sizeof(R) * (size_t) d);
  ths->sigma = (R*) PNX(malloc)(sizeof(R) * (size_t) d);
  for(int t = 0;t < d; t++)
    ths->sigma[t] = ((R)n[t])/N[t];


  /* Precompute Flags not yet implemented */
  if(nfft_flags & (PNFFT_PRE_ONE_PSI| PNFFT_PRE_PHI_HUT)){
    PX(printf)(comm_cart, "!!! Warning: Precompute Flags are not yet implemented in PNFFT !!!\n");
    ths->nfft_flags &= (~PNFFT_PRE_ONE_PSI); /* needed for correct nfft_finalize */
    ths->nfft_flags &= (~PNFFT_PRE_PHI_HUT); /* needed for correct nfft_finalize */
  }

  /* init window specific parameters */
  ths->exp_const = NULL; /* needed for correct nfft_finalize */
  ths->b = (R*) PNX(malloc)(sizeof(R) * (size_t) d);
  if(nfft_flags & PNFFT_WINDOW_GAUSSIAN){
    for(int t=0; t<d; t++)
      ths->b[t]= 2*ths->sigma[t] / (2*ths->sigma[t]-1) * ((R)m / PNFFT_PI);
    ths->exp_const = (R*) PNX(malloc)(sizeof(R) * (size_t) d * ths->cutoff);
    for(int t=0; t<d; t++)
      for(int s=0; s<ths->cutoff; s++)
        ths->exp_const[ths->cutoff*t+s] = exp(-s*s/ths->b[t])/(sqrt(PNFFT_PI*ths->b[t]));
  } else { /* default window function is Kaiser-Bessel */
    for(int t=0; t<d; t++)
      ths->b[t] = (double) PNFFT_PI * (2.0 - 1.0/ths->sigma[t]);
  }

  get_size_gcells(m, ths->cutoff,
      gcells_below, gcells_above);

  alloc_local_data_in = PX(local_size_many_dft)(3, n, N, no, howmany,
      PFFT_DEFAULT_BLOCKS, PFFT_DEFAULT_BLOCKS, comm_cart, PFFT_TRANSPOSED_IN,
      local_N, local_N_start, local_no, local_no_start);

  alloc_local_gc = PX(local_size_many_gc)(3, local_no, local_no_start,
      alloc_local_data_in, howmany, gcells_below, gcells_above,
      local_ngc, local_gc_start);
  
  alloc_local_f_hat = PNX(prod_INT)(3, local_N);
  if(nfft_flags & PNFFT_MALLOC_F_HAT)
    ths->f_hat = (C*) PNX(malloc)(sizeof(C) * (size_t) alloc_local_f_hat);

  if(nfft_flags & PNFFT_FFT_INIT){
    ths->g2 = (C*) PNX(malloc)(sizeof(C) * (size_t) alloc_local_gc);
    if(nfft_flags & PNFFT_FFT_OUT_OF_PLACE)
      ths->g1 = (C*) PNX(malloc)(sizeof(C) * (size_t) alloc_local_data_in);
    else
      ths->g1 = ths->g2;
    
    ths->pfft_forw = PX(plan_many_dft)(3, n, N, no, howmany,
        PFFT_DEFAULT_BLOCKS, PFFT_DEFAULT_BLOCKS, ths->g1, ths->g2, comm_cart,
        FFTW_FORWARD, PFFT_TRANSPOSED_IN, fftw_flags);
    
    ths->pfft_back = PX(plan_many_dft)(3, n, no, N, howmany,
        PFFT_DEFAULT_BLOCKS, PFFT_DEFAULT_BLOCKS, ths->g2, ths->g1, comm_cart,
        FFTW_BACKWARD, PFFT_TRANSPOSED_OUT, fftw_flags);

    ths->gcplan = PX(plan_many_cgc)(3, no, howmany, PFFT_DEFAULT_BLOCKS,
        gcells_below, gcells_above, ths->g2, comm_cart, 0);
  } else {
    ths->g1 = ths->g2 = NULL;
    ths->pfft_forw = ths->pfft_back = NULL;
    ths->gcplan = NULL;
  }

  /* init compute flags before malloc f and f_grad */
  ths->compute_flags = 0;

  /* allocate mem and adjust nfft_flags, compute_flags */
  PNX(malloc_x)(ths, nfft_flags);
  PNX(malloc_f)(ths, nfft_flags);
  PNX(malloc_grad_f)(ths, nfft_flags);
  PNX(reset_timer)(ths);
}

void PNX(malloc_x)(
    PNX(plan) *ths, unsigned nfft_flags
    )
{
  ths->x = NULL;
  if( (nfft_flags & PNFFT_MALLOC_X) && (ths->local_M > 0) ){
    ths->x = (R*) PNX(malloc)(sizeof(R) * (size_t) ths->d*ths->local_M);
    ths->nfft_flags |= PNFFT_MALLOC_X;
  }
}

void PNX(free_x)(
    PNX(plan) *ths, unsigned nfft_finalize_flags
    )
{
  if( (nfft_finalize_flags & PNFFT_FREE_X) && (ths->x != NULL) ){
    PNX(free)(ths->x);
    ths->nfft_flags &= ~PNFFT_MALLOC_X;
  }
}

void PNX(malloc_f)(
    PNX(plan) *ths, unsigned nfft_flags
    )
{
  ths->f = NULL;
  if( (nfft_flags & PNFFT_MALLOC_F) && (ths->local_M > 0) ){
    ths->f = (C*) PNX(malloc)(sizeof(C) * (size_t) ths->local_M);
    ths->nfft_flags |= PNFFT_COMPUTE_F;
    ths->compute_flags |= PNFFT_COMPUTE_F;
  }
}

void PNX(free_f)(
    PNX(plan) *ths, unsigned nfft_finalize_flags
    )
{
  if( (nfft_finalize_flags & PNFFT_FREE_F) && (ths->f != NULL) ){
    PNX(free)(ths->f);
    ths->nfft_flags &= ~PNFFT_MALLOC_F;
    ths->compute_flags &= ~PNFFT_COMPUTE_F;
  }
}

void PNX(malloc_grad_f)(
    PNX(plan) *ths, unsigned nfft_flags
    )
{
  ths->grad_f = NULL;
  if( (nfft_flags & PNFFT_MALLOC_GRAD_F) && (ths->local_M > 0) ){
    ths->grad_f = (C*) PNX(malloc)(sizeof(C) * (size_t) ths->d*ths->local_M);
    ths->nfft_flags |= PNFFT_MALLOC_GRAD_F;
    ths->compute_flags |= PNFFT_COMPUTE_GRAD_F;
  }
}

void PNX(free_grad_f)(
    PNX(plan) *ths, unsigned nfft_finalize_flags
    )
{
  if( (nfft_finalize_flags & PNFFT_FREE_GRAD_F) && (ths->grad_f != NULL) ){
    PNX(free)(ths->grad_f);
    ths->nfft_flags &= ~PNFFT_MALLOC_GRAD_F;
    ths->compute_flags &= ~PNFFT_COMPUTE_GRAD_F;
  }
}


void PNX(trafo_F)(
    PNX(plan) *ths
    )
{
  PX(execute)(ths->pfft_forw);
}

void PNX(adjoint_F)(
    PNX(plan) *ths
    )
{
  PX(execute)(ths->pfft_back);
}


static void fftshift_N_half(
    const INT *N, const INT *n, const INT *no,
    const INT *local_no, const INT *local_no_start,
    int sign,
    C *g
    )
{
  INT l0, l1, l2, m=0;
 
  for(l0=local_no_start[0]; l0<local_no[0]+local_no_start[0]; l0++)
    for(l1=local_no_start[1]; l1<local_no[1]+local_no_start[1]; l1++)
      for(l2=local_no_start[2]; l2<local_no[2]+local_no_start[2]; l2++, m++)
        g[m] *= cexp(-sign * PNFFT_PI * I * ( N[0]*l0/((R)n[0]) + N[1]*l1/((R) n[1]) + N[2]*l2/((R) n[2]) ));
}


void PNX(trafo_B)(
    PNX(plan) *ths
    )
{
  const int cutoff = ths->cutoff;
  INT j, l0, l1, l2, m0, m1, m2;
  INT *sorted_index = NULL;
  INT local_no[3], local_no_start[3];
  INT gcells_below[3], gcells_above[3];
  INT u_j[3], local_ngc[3];
  R *pre_psi = NULL;
  R *pre_psi_derive = NULL;
  
  PNX(local_size_B)(ths->N, ths->n, ths->no, ths->comm_cart,
      local_no, local_no_start);

  fftshift_N_half(ths->N, ths->n, ths->no, local_no, local_no_start, FFTW_FORWARD,
      ths->g2);

  get_size_gcells(ths->m, ths->cutoff,
      gcells_below, gcells_above);
  local_array_size(local_no, gcells_below, gcells_above,
      local_ngc);

  /* send ghost cells in ring */
  ths->timer_trafo.gcells -= MPI_Wtime();
  PX(exchange)(ths->gcplan);
  ths->timer_trafo.gcells += MPI_Wtime();

  /* sort indices for better cache handling */
  if(ths->nfft_flags & PNFFT_SORT_NODES){
    ths->timer_trafo.sort_nodes -= MPI_Wtime();
    sorted_index = (INT*) PNX(malloc)(sizeof(INT) * (size_t) 2*ths->local_M);
    sort_nodes_for_better_cache_handle(
        ths->n, ths->m, ths->local_M, ths->x,
        sorted_index);
    ths->timer_trafo.sort_nodes += MPI_Wtime();
  }

  ths->timer_trafo.loop_B -= MPI_Wtime();
  pre_psi = (R*) PNX(malloc)(sizeof(R) * (size_t) cutoff*3);
  if(ths->compute_flags & PNFFT_COMPUTE_GRAD_F)
    pre_psi_derive = (R*) PNX(malloc)(sizeof(R) * (size_t) cutoff*3);
  for(INT p=0; p<ths->local_M; p++){
    j = (ths->nfft_flags & PNFFT_SORT_NODES) ? sorted_index[2*p+1] : p;
    
    /* evaluate window on axes */
    pre_psi_tensor(
        ths->n, ths->b, ths->m, ths->cutoff, &(ths->x[ths->d*j]), ths->exp_const, ths->nfft_flags,
        pre_psi);
  
    if(ths->compute_flags & PNFFT_COMPUTE_GRAD_F)
      pre_psi_derive_tensor(
          ths->n, ths->b, ths->m, ths->cutoff, &(ths->x[ths->d*j]), pre_psi, ths->nfft_flags,
          pre_psi_derive);

    lowest_summation_index(
        ths->n, ths->m, &(ths->x[ths->d*j]), local_no_start, gcells_below,
        u_j);
    
    /* compute f */
    if(ths->compute_flags & PNFFT_COMPUTE_F){
      ths->f[j] = 0;
      m0 = MACRO_PLAIN_INDEX_3D(u_j, local_ngc);
      for(l0=0; l0<cutoff; l0++, m0 += local_ngc[1]*local_ngc[2])
        for(l1=0, m1=m0; l1<cutoff; l1++, m1 += local_ngc[2])
          for(l2=0, m2 = m1; l2<cutoff; l2++, m2++)
            ths->f[j] += pre_psi[l0] * pre_psi[cutoff+l1] * pre_psi[2*cutoff+l2] * ths->g2[m2];
    }

    /* compute grad_f */
    if(ths->compute_flags & PNFFT_COMPUTE_GRAD_F){
      for(int t=0; t<ths->d; t++)
        ths->grad_f[ths->d*j+t] = 0;
      m0 = MACRO_PLAIN_INDEX_3D(u_j, local_ngc);
      for(l0=0; l0<cutoff; l0++, m0 += local_ngc[1]*local_ngc[2]){
        for(l1=0, m1=m0; l1<cutoff; l1++, m1 += local_ngc[2]){
          for(l2=0, m2 = m1; l2<cutoff; l2++, m2++){
            ths->grad_f[ths->d*j+0] += pre_psi_derive[l0] * pre_psi[cutoff+l1] * pre_psi[2*cutoff+l2] * ths->g2[m2];
            ths->grad_f[ths->d*j+1] += pre_psi[l0] * pre_psi_derive[cutoff+l1] * pre_psi[2*cutoff+l2] * ths->g2[m2];
            ths->grad_f[ths->d*j+2] += pre_psi[l0] * pre_psi[cutoff+l1] * pre_psi_derive[2*cutoff+l2] * ths->g2[m2];
          }
        }
      }
    }
  }
  ths->timer_trafo.loop_B += MPI_Wtime();
  
  if(ths->nfft_flags & PNFFT_SORT_NODES)
    PNX(free)(sorted_index);

  PNX(free)(pre_psi);
  if(ths->compute_flags & PNFFT_COMPUTE_GRAD_F)
    PNX(free)(pre_psi_derive);
}


void PNX(adjoint_B)(
    PNX(plan) *ths
    )
{
  const int cutoff = ths->cutoff;
  INT j, l0, l1, l2, m0, m1, m2;
  INT *sorted_index = NULL;
  INT local_no[3], local_no_start[3];
  INT gcells_below[3], gcells_above[3];
  INT u_j[3], local_ngc[3], local_ngc_total;
  R *pre_psi = NULL;
  
  PNX(local_size_B)(ths->N, ths->n, ths->no, ths->comm_cart,
      local_no, local_no_start);
  
  get_size_gcells(ths->m, ths->cutoff,
      gcells_below, gcells_above);
  local_array_size(local_no, gcells_below, gcells_above,
      local_ngc);

  local_ngc_total = PNX(prod_INT)(3, local_ngc);
  memset(ths->g2, 0, local_ngc_total * sizeof(C));

  /* sort indices for better cache handling */
  if(ths->nfft_flags & PNFFT_SORT_NODES){
    ths->timer_adj.sort_nodes -= MPI_Wtime();
    sorted_index = (INT*) PNX(malloc)(sizeof(INT) * (size_t) 2*ths->local_M);
    sort_nodes_for_better_cache_handle(
        ths->n, ths->m, ths->local_M, ths->x,
        sorted_index);
    ths->timer_adj.sort_nodes += MPI_Wtime();
  }

  ths->timer_adj.loop_B -= MPI_Wtime();
  pre_psi = (R*) PNX(malloc)(sizeof(R) * (size_t) cutoff*3);
  for(INT p=0; p<ths->local_M; p++){
    j = (ths->nfft_flags & PNFFT_SORT_NODES) ? sorted_index[2*p+1] : p;

    /* evaluate window on axes */
    pre_psi_tensor(
        ths->n, ths->b, ths->m, ths->cutoff, &(ths->x[ths->d*j]), ths->exp_const, ths->nfft_flags,
        pre_psi);

    lowest_summation_index(
        ths->n, ths->m, &(ths->x[ths->d*j]), local_no_start, gcells_below,
        u_j);

    m0 = MACRO_PLAIN_INDEX_3D(u_j, local_ngc);
    for(l0=0; l0<cutoff; l0++, m0 += local_ngc[1]*local_ngc[2])
      for(l1=0, m1=m0; l1<cutoff; l1++, m1 += local_ngc[2])
        for(l2=0, m2 = m1; l2<cutoff; l2++, m2++ )
           ths->g2[m2] += pre_psi[l0] * pre_psi[cutoff+l1] * pre_psi[2*cutoff+l2] * ths->f[j];
  }
  ths->timer_adj.loop_B += MPI_Wtime();
  
  /* reduce ghost cells in ring */
  ths->timer_adj.gcells -= MPI_Wtime();
  PX(reduce)(ths->gcplan);
  ths->timer_adj.gcells += MPI_Wtime();
 
  fftshift_N_half(ths->N, ths->n, ths->no, local_no, local_no_start, FFTW_BACKWARD,
      ths->g2);
  
  if(ths->nfft_flags & PNFFT_SORT_NODES)
    PNX(free)(sorted_index);
  PNX(free)(pre_psi);
}


/* Implement ghostcell send for all dimensions */
static void get_size_gcells(
    int m, int cutoff,
    INT *gcells_below, INT *gcells_above
    )
{
  for(int t=0; t<3; t++){
    gcells_below[t] = m;
    gcells_above[t] = cutoff - gcells_below[t] - 1;
  }
}

static void lowest_summation_index(
    const INT *n, int m, const R *x,
    const INT *local_no_start, const INT *gcells_below,
    INT *u_j
    )
{
  project_node_to_grid(n, m, x, u_j);
  for(int t=0; t<3; t++)
    u_j[t] = u_j[t] - local_no_start[t] + gcells_below[t];
}


static void local_array_size(
    const INT *local_n, const INT *gcells_below, const INT *gcells_above,
    INT *local_ngc
    )
{
  for(int t=0; t<3; t++)
    local_ngc[t] = local_n[t] + gcells_below[t] + gcells_above[t];
}

/* calculate window function */
static void pre_psi_tensor(
    const INT *n, const R *b, int m, int cutoff,
    const R *x, const R *exp_const, unsigned nfft_flags,
    R *pre_psi
    )
{
  if(nfft_flags & PNFFT_WINDOW_GAUSSIAN)
    pre_psi_tensor_gaussian(
        n, b, m, cutoff, x, exp_const,
        pre_psi);
  else
    pre_psi_tensor_kaiser_bessel(
        n, b, m, cutoff, x,
        pre_psi);
}

static void pre_psi_tensor_kaiser_bessel(
    const INT *n, const R *b, int m, int cutoff,
    const R *x,
    R *pre_psi
    )
{
  const int d=3;
  R u_j;

  for(int t=0; t<d; t++){
    u_j = floor( n[t]*x[t] - m );
    for(int s=0; s<cutoff; s++)
      pre_psi[cutoff*t+s] = kaiser_bessel_1d(
          (u_j + s) / n[t] - x[t],
          n[t], b[t], m);
  }
}

static void pre_psi_tensor_gaussian(
    const INT *n, const R *b, int m, int cutoff,
    const R *x, const R *exp_const,
    R *fg_psi
    )
{
  const int d=3;
  R u_j, exp_sqr, exp_lin, tmp;

  for(int t=0; t<d; t++){
    u_j = floor( n[t]*x[t] - m );
    exp_sqr = exp( -SQR( n[t]*x[t]-u_j ) / b[t] );
    exp_lin = exp( 2*( n[t]*x[t]-u_j ) / b[t] );

    tmp = exp_sqr;
    for(int s=0; s<cutoff; s++){
      fg_psi[cutoff*t+s] = tmp * exp_const[cutoff*t+s];
      tmp *= exp_lin;
    }
  }
}

/* calculate window derivative */
static void pre_psi_derive_tensor(
    const INT *n, const R *b, int m, int cutoff,
    const R *x, const R *pre_psi, unsigned nfft_flags,
    R *pre_psi_derive
    )
{
  if(nfft_flags & PNFFT_WINDOW_GAUSSIAN)
    pre_psi_derive_tensor_gaussian(
        n, b, m, cutoff, x, pre_psi,
        pre_psi_derive);
  else
    pre_psi_derive_tensor_kaiser_bessel(
        n, b, m, cutoff, x, pre_psi,
        pre_psi_derive);
}

static void pre_psi_derive_tensor_kaiser_bessel(
    const INT *n, const R *b, int m, int cutoff,
    const R *x, const R *pre_psi,
    R *pre_psi_derive
    )
{
  const int d=3;
  R u_j;

  for(int t=0; t<d; t++){
    u_j = floor( n[t]*x[t] - m );
    for(int s=0; s<cutoff; s++)
      pre_psi_derive[cutoff*t+s] = kaiser_bessel_derivative_1d(
          (u_j + s) / n[t] - x[t],
          n[t], b[t], m, pre_psi[cutoff*t+s]);
  }
}

static void pre_psi_derive_tensor_gaussian(
    const INT *n, const R *b, int m, int cutoff,
    const R *x, const R *fg_psi,
    R *fg_psi_derive
    )
{
  const int d=3;
  R u_j;

  for(int t=0; t<d; t++){
    u_j = n[t]*x[t] - floor( n[t]*x[t] ) + m;
    for(int s=0; s<cutoff; s++)
      fg_psi_derive[cutoff*t+s] = -2.0*n[t]/b[t] * (u_j-s) * fg_psi[cutoff*t+s];
  }
}

static void sort_nodes_for_better_cache_handle(
    const INT *n, int m, INT local_M, const R *x,
    INT *sort
    )
{ /* sort must be of length 2*local_M */
  int d = 3;
  INT u_j[3];

  if(local_M == 0)
    return;

  for(INT k = 0; k < local_M; k++){
    project_node_to_grid(n, m, x+d*k, u_j);
    sort[2*k]   = MACRO_PLAIN_INDEX_3D(u_j, n);
    sort[2*k+1] = k;
  }
  qsort(sort, (size_t) local_M, 2*sizeof(INT), compare_INT);
}

static void project_node_to_grid(
    const INT *n, int m, const R *x,
    INT *u_j
    )
{
  for(int t=0; t<3; t++)
    u_j[t] = lrint(floor(n[t]*x[t])) - m;
}

static void get_mpi_cart_dims(
    MPI_Comm comm_cart,
    int *dims
    )
{
  int maxdims, *periods, *coords;
 
  MPI_Cartdim_get(comm_cart, &maxdims);

  periods = (int*) malloc(sizeof(int) * (size_t) maxdims);
  coords  = (int*) malloc(sizeof(int) * (size_t) maxdims);
  MPI_Cart_get(comm_cart, maxdims, dims, periods, coords);
  free(periods);
  free(coords);
 
  if(maxdims == 2)
    dims[2] = 1;
}

static int compare_INT(
    const void* a, const void* b
    )
{
  return (*(const INT*)a <= *(const INT*)b) ? -1 : 1;
}


static R kaiser_bessel_1d(
    R x, INT n, R b, int m
    )
{
  /* TODO: try to avoid case d<0, since in theory d >= 0 */
  R d, r;

  d = SQR( (R)m ) - SQR( x*n );
  r = (d<0) ? sqrt(-d) : sqrt(d);

  if(d < 0)
    return sin(b*r) / ((R)PNFFT_PI*r);

  /* regular case d >= 0 */
  return (d>0) ? sinh(b*r) / ((R)PNFFT_PI*r) : b/PNFFT_PI;
}

// static R kaiser_bessel_derivative_1d_vers1(
//     R x, INT n, R b, int m
//     )
// {
//   /* TODO: try to avoid case d<0, since in theory d >= 0 */
//   R d, r;
// 
//   d = SQR( (R)m ) - SQR( x*n );
//   r = (d<0) ? sqrt(-d) : sqrt(d);
// 
//   if(d < 0) /* use of -d results in change of sign within the brackets */
//     return n*n*x/(PNFFT_PI*d) * (sin(b*r)/r - b*cos(b*r));
// 
//   /* regular case d >= 0 */
//   return (d>0) ? n*n*x/(PNFFT_PI*d) * (sinh(b*r)/r - b*cosh(b*r)) : -n*m*b*b*b/(3.0*PNFFT_PI);
// }

static R kaiser_bessel_derivative_1d(
    R x, INT n, R b, int m, R psi
    )
{
  /* TODO: try to avoid case d<0, since in theory d >= 0 */
  R d, r;

  d = SQR( (R)m ) - SQR( x*n );
  r = (d<0) ? sqrt(-d) : sqrt(d);

  if(d < 0) /* use of -d results in change of sign within the brackets */
    return n*n*x/d * (b*cos(b*r)/PNFFT_PI - psi);

  /* regular case d >= 0 */
  return (d>0) ? n*n*x/d * (b*cosh(b*r)/PNFFT_PI - psi) : -n*m*b*b*b/(3.0*PNFFT_PI);
}



