/*
 * Copyright (c) 2011 Michael Pippig
 *
 * This file is part of PNFFT.
 *
 * PNFFT is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * PNFFT is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with PNFFT.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include "pnfft.h"
#include "ipnfft.h"

static void reset_timer(
    PNX(timer) *timer);
static int file_exists(
    const char *name); 
static FILE* open_or_create_file_to_append(
    MPI_Comm comm, const char *name);
static void write_info_header(
    MPI_Comm comm, FILE *file);
static int get_matlab_index(
    MPI_Comm comm);
static void write_run_specific_infos(
    MPI_Comm comm, FILE *file, PNX(plan) *ths);
static void fprint_average_timer_internal(
    MPI_Comm comm, FILE *file, const char *prefix, PNX(timer) *ths,
    unsigned flags);
static void fprint_average_timer(
    MPI_Comm comm, FILE *file, PNX(plan) *ths, unsigned flags);

void PNX(get_timer_trafo)(
    PNX(plan) *ths, PNX(timer) *timer
    )
{
  PNX(copy_timer)(&ths->timer_trafo, timer);
}

void PNX(get_timer_adj)(
    PNX(plan) *ths, PNX(timer) *timer
    )
{
  PNX(copy_timer)(&ths->timer_adj, timer);
}


void PNX(convert_timer2vec)(
    const PNX(timer) *timer,
    double *times
    )
{
  times[0] = (double) timer->iter;
  times[1] = timer->whole;
  times[2] = timer->loop_B;
  times[3] = timer->sort_nodes;
  times[4] = timer->gcells;
  times[5] = timer->matrix_B;
  times[6] = timer->matrix_F;
  times[7] = timer->matrix_D;
}


void PNX(convert_vec2timer)(
    const double *times,
    PNX(timer) *timer
    )
{
  timer->iter = round(times[0]);
  timer->whole = times[1];
  timer->loop_B = times[2];
  timer->sort_nodes = times[3];
  timer->gcells = times[4];
  timer->matrix_B = times[5];
  timer->matrix_F = times[6];
  timer->matrix_D = times[7];
}

void PNX(average_timer)(
     PNX(timer) *timer
     )
{
  double times[8];
  
  if(timer->iter == 0)
    return;
  
  PNX(convert_timer2vec)(timer, times);
  for(int t=1; t<8; t++)
    times[t] /= times[0];
  PNX(convert_vec2timer)(times, timer);
}

void PNX(copy_timer)(
    const PNX(timer) *orig,
    PNX(timer) *copy
    )
{
  double times[8];
  
  PNX(convert_timer2vec)(orig, times);
  PNX(convert_vec2timer)(times, copy);
}

void PNX(reduce_max_timer)(
    MPI_Comm comm, const PNX(timer) *timer,
    PNX(timer) *timer_max
    )
{
  double times[8], times_max[8];
  
  PNX(convert_timer2vec)(timer, times);
  MPI_Reduce(times, times_max, 8, MPI_DOUBLE, MPI_MAX, 0, comm);
  PNX(convert_vec2timer)(times_max, timer_max);
}

void PNX(add_timers)(
    const PNX(timer) *sum1, const PNX(timer) *sum2,
    PNX(timer) *res
    )
{
  double t1[8], t2[8], sum[8];
  
  PNX(convert_timer2vec)(sum1, t1);
  PNX(convert_timer2vec)(sum2, t2);
  for(int t=0; t<8; t++)
    sum[t] = t1[t] + t2[t];
  PNX(convert_vec2timer)(sum, res);
}


void PNX(reset_timer)(
    PNX(plan) *ths
    )
{
  reset_timer(&ths->timer_trafo);
  reset_timer(&ths->timer_adj);
}

static void reset_timer(
    PNX(timer) *timer
    )
{
  double times[8];
  
  for(int t=0; t<8; t++)
    times[t] = 0;
  PNX(convert_vec2timer)(times, timer);
}


void PNX(write_average_timer)(
    PNX(plan) *ths, const char *name, MPI_Comm comm
    )
{
  int newfile;
  FILE *f;
  
  newfile = !file_exists(name);
  f = open_or_create_file_to_append(comm, name);
  if( newfile )
    write_info_header(comm, f);
  
  write_run_specific_infos(comm, f, ths);
  fprint_average_timer(comm, f, ths, PNFFT_PRINT_TIMER_BASIC);
  
  fclose(f);
}

void PNX(write_average_timer_adv)(
    PNX(plan) *ths, const char *name, MPI_Comm comm
    )
{
  FILE *f;
  PNX(write_average_timer)(ths, name, comm);
  f = open_or_create_file_to_append(comm, name);
  fprint_average_timer(comm, f, ths, PNFFT_PRINT_TIMER_ADV);
  fclose(f);
  PX(write_average_timer_adv)(ths->pfft_forw, name, comm);
  PX(write_average_timer_adv)(ths->pfft_back, name, comm);
  PX(write_average_gctimer_adv)(ths->gcplan, name, comm);
}

void PNX(print_average_timer)(
    PNX(plan) *ths, MPI_Comm comm
    )
{
  write_info_header(comm, stdout);
  write_run_specific_infos(comm, stdout, ths);
  fprint_average_timer(comm, stdout, ths, PNFFT_PRINT_TIMER_BASIC);
}

void PNX(print_average_timer_adv)(
    PNX(plan) *ths, MPI_Comm comm
    )
{
  PNX(print_average_timer)(ths, comm);
  fprint_average_timer(comm, stdout, ths, PNFFT_PRINT_TIMER_ADV);
  PX(print_average_timer_adv)(ths->pfft_forw, comm);
  PX(print_average_timer_adv)(ths->pfft_back, comm);
  PX(print_average_gctimer_adv)(ths->gcplan, comm);
}



static int file_exists(
    const char *name
    )
{
  FILE *f;
  
  f=fopen(name, "r");
  if(f != NULL)
    fclose(f);
  
  return (f != NULL);
}

static FILE* open_or_create_file_to_append(
    MPI_Comm comm, const char *name
    )
{
  FILE *f=fopen(name, "a+");
  if(f==NULL){
    PX(fprintf)(comm, stderr, "Error: Cannot open file %s.\n", name);
    exit(1);
  }
  return f;
}

static void write_info_header(
    MPI_Comm comm, FILE *file
    )
{
  PX(fprintf)(comm, file, "%% N  - NFFT size\n");
  PX(fprintf)(comm, file, "%% n  - FFT size\n");
  PX(fprintf)(comm, file, "%% np - process grid\n");
  PX(fprintf)(comm, file, "%% procs - number of processes\n");
  PX(fprintf)(comm, file, "%% pnfft - PNFFT runtime\n");
  PX(fprintf)(comm, file, "%% pfft  - PFFT runtime\n");
  PX(fprintf)(comm, file, "%% index(i) = log(procs(i)) + 1\n");
}

static int get_matlab_index(
    MPI_Comm comm
    )
{
  int size;
  MPI_Comm_size(comm, &size);
  return lrint(round(log((R) size)/log(2.0)))+1;
}

static void write_run_specific_infos(
    MPI_Comm comm, FILE *file, PNX(plan) *ths
    )
{
  
  int size;
  int idx = get_matlab_index(comm);
  MPI_Comm_size(comm, &size);
  
  PX(fprintf)(comm, file, "\nindex(%d) = %d;  ", idx, idx);
  PX(fprintf)(comm, file, "procs(%d) = %d;  ", idx, size);
  PX(fprintf)(comm, file, "np_pnfft(%d, 1:3) = [%d %d %d];  ", idx, ths->np[0], ths->np[1], ths->np[2]);
  PX(fprintf)(comm, file, "N_pnfft(%d, 1:%d) = [", idx, ths->d);
  for(int t=0; t<ths->d; t++)
    PX(fprintf)(comm, file, "%td ", ths->N[t]);
  PX(fprintf)(comm, file, "];\n");
}

static void fprint_average_timer_internal(
    MPI_Comm comm, FILE *file, const char *prefix, PNX(timer) *ths,
    unsigned flags
    )
{
  int idx = get_matlab_index(comm);
  PNX(timer) mt;
  PNX(reduce_max_timer)(comm, ths, &mt);
  PNX(average_timer)(&mt);

  if(flags & PNFFT_PRINT_TIMER_BASIC){
    PX(fprintf)(comm, file, "%s_iter(%d)    = %d;  ", prefix, idx, ths->iter);
    PX(fprintf)(comm, file, "%s(%d)   = %.3e;\n", prefix, idx, mt.whole);
  } else if(flags & PNFFT_PRINT_TIMER_ADV){
    PX(fprintf)(comm, file, "%s_matrix_D(%d)   = %.3e;  ", prefix, idx, mt.matrix_D);
    PX(fprintf)(comm, file, "%s_matrix_F(%d)   = %.3e;\n", prefix, idx, mt.matrix_F);
    PX(fprintf)(comm, file, "%s_matrix_B(%d)   = %.3e;  ", prefix, idx, mt.matrix_B);
    PX(fprintf)(comm, file, "%s_gcells(%d)     = %.3e;\n", prefix, idx, mt.gcells);
    PX(fprintf)(comm, file, "%s_sort_nodes(%d) = %.3e;  ", prefix, idx, mt.sort_nodes);
    PX(fprintf)(comm, file, "%s_loop_B(%d)     = %.3e;\n", prefix, idx, mt.loop_B);
  }
}

static void fprint_average_timer(
    MPI_Comm comm, FILE *file, PNX(plan) *ths, unsigned flags
    )
{
  fprint_average_timer_internal(comm, file, "pnfft_trf", &ths->timer_trafo, flags);
  fprint_average_timer_internal(comm, file, "pnfft_adj", &ths->timer_adj, flags);
}
  


